# type: ignore
import sys

if sys.version_info >= (3,10,0):
    from importlib import metadata

    def get_entry_points(group_name):
        return metadata.entry_points().select(group=group_name)

elif sys.version_info >= (3,8,0):
    from importlib import metadata

    def get_entry_points(group_name):
        return metadata.entry_points().get(group_name, tuple())

else:
    import pkg_resources
    def get_entry_points(group_name):
        return pkg_resources.iter_entry_points(group_name)



from typing import TYPE_CHECKING, Any, List

from . import entry_points
from ..importer import Importer

if TYPE_CHECKING:
    import argparse
    from systemrdl import RDLCompiler

class ImporterPluginWrapper(Importer):
    """
    Importers external to this package can register an implementation that can
    be loaded into PeakRDL

    The imporer definition is provided by a class that mimics the
    members/methods of Importer without actually extending it

    .. code:: python

        class MyImporter:
            file_extensions = ["foo"]

            def is_compatible(self, path: str) -> bool:
                raise NotImplementedError

            def add_importer_arguments(self, arg_group: 'argparse.ArgumentParser') -> None:
                pass

            def do_import(self, rdlc: 'RDLCompiler', options: 'argparse.Namespace', path: str):
                raise NotImplementedError
    """

    def __init__(self, name: str, importer_plugin_cls: Any) -> None:
        self.plugin = importer_plugin_cls()
        self.name = name
        self.file_extensions = getattr(self.plugin, "file_extensions", [])

    def __repr__(self) -> str:
        return "<%s '%s' from %s at 0x%x>" % (
            self.__class__.__qualname__,
            self.name,
            self.plugin.__class__,
            id(self)
        )

    def is_compatible(self, path: str) -> bool:
        func = getattr(self.plugin, "is_compatible", None)
        if callable(func):
            return func(path)
        else:
            raise NotImplementedError


    def add_importer_arguments(self, arg_group: 'argparse._ActionsContainer') -> None:
        func = getattr(self.plugin, "add_importer_arguments", None)
        if callable(func):
            func(arg_group)


    def do_import(self, rdlc: 'RDLCompiler', options: 'argparse.Namespace', path: str) -> None:
        func = getattr(self.plugin, "do_import", None)
        if callable(func):
            func(rdlc, options, path)
        else:
            raise NotImplementedError



def get_importer_plugins() -> List[Importer]:
    """
    Load any plugins that advertise themselves in their setup.py via the following:

    setup(
        ...
        entry_points = {
            "peakrdl.importers": [
                'my_importer_name = module.path.to:MyImporter'
            ]
        },
    )
    """
    importers = []
    for ep in entry_points.get_entry_points("peakrdl.importers"): # type: ignore
        importer = ImporterPluginWrapper(ep.name, ep.load())
        importers.append(importer)

    return importers # type: ignore



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List
    import argparse
    from systemrdl import RDLCompiler

class Importer:
    #: Importer name

    name = None # type: str

    #: List of possible file extensions.
    #: This is used as the first pass to determine if the input file shall use
    #: this importer
    file_extensions = [] # type: List[str]

    def is_compatible(self, path: str) -> bool:
        """
        If the file extension was not enough to determine which importer to use,
        this function is called and shall determine if the file is compatible
        with this importer to a high degree of confidence.

        Note: this should not attempt to validate the file's correctness, just
        perform a low-cost assessment if the file appears to match the format
        that this importer supports.
        """
        raise NotImplementedError


    def add_importer_arguments(self, arg_group: 'argparse._ActionsContainer') -> None:
        pass


    def do_import(self, rdlc: 'RDLCompiler', options: 'argparse.Namespace', path: str) -> None:
        raise NotImplementedError



import argparse
import sys
from typing import TYPE_CHECKING, List, Dict

from systemrdl import RDLCompileError

from .__about__ import __version__
from .plugins.exporter import get_exporter_plugins
from .cmd.dump import Dump
from .cmd.list_globals import ListGlobals


if TYPE_CHECKING:
    from .subcommand import Subcommand


DESCRIPTION = """
PeakRDL is a control & status register model automation toolchain.

For help about a specific subcommand, try:
    peakrdl <command> --help

For more documentation, visit https://peakrdl.readthedocs.io

"""


class SubcommandHelpFormatter(argparse.RawDescriptionHelpFormatter):
    def _format_action(self, action): # type: ignore
        parts = super(argparse.RawDescriptionHelpFormatter, self)._format_action(action)
        if action.nargs == argparse.PARSER:
            parts = "\n".join(parts.split("\n")[1:])
        return parts


def main() -> None:
    # Collect all subcommands
    subcommands = [
        Dump(),
        ListGlobals(),
    ] # type: List[Subcommand]
    subcommands += get_exporter_plugins()

    # Check for duplicates
    sc_dict = {} # type: Dict[str, Subcommand]
    for sc in subcommands:
        if sc.name in sc_dict:
            raise RuntimeError(f"More than one exporter plugin was registered with the same name '{sc.name}': \n\t{sc_dict[sc.name]}\n\t{sc}")
        sc_dict[sc.name] = sc

    # Initialize top-level arg parser
    parser = argparse.ArgumentParser(
        description=DESCRIPTION,
        formatter_class=SubcommandHelpFormatter,
    )
    parser.add_argument("--version", action="version", version=__version__)

    # Initialize subcommand arg parsers
    subgroup = parser.add_subparsers(
        title="subcommands",
        metavar="<subcommand>",
        required=True
    )
    for subcommand in subcommands:
        subcommand._init_subparser(subgroup)

    # Execute!
    options = parser.parse_args()
    try:
        options.subcommand.main(options)
    except RDLCompileError:
        sys.exit(1)



from typing import TYPE_CHECKING, Optional
import argparse
import os
import sys
import shlex

from systemrdl import RDLCompiler

from .plugins.importer import get_importer_plugins

from . import process_input

if TYPE_CHECKING:
    from systemrdl.node import AddrmapNode


class LoadArgsFromFile (argparse.Action):
    def __call__ (self, parser, namespace, values, option_string = None): # type: ignore
        argfile = values

        if not os.path.exists(argfile):
            print(f"file not found: {argfile}", file=sys.stderr)
            sys.exit(1)

        with open(argfile, "r", encoding='utf-8') as f:
            parser.parse_args(
                shlex.split(f.read(), comments=True),
                namespace
            )


class Subcommand:
    """
    Base command line interface subcommand class
    """

    #: Subcommand name
    name = None # type: str

    #: Short-form description
    short_desc = None # type: str

    #: Longer-form description
    #: If left as None, inherits short_desc
    long_desc = None # type: Optional[str]


    def _init_subparser(self, subgroup: 'argparse._SubParsersAction') -> None:
        assert isinstance(self.name, str)
        assert isinstance(self.short_desc, str)
        subparser = subgroup.add_parser(
            self.name,
            help=self.short_desc,
            description=(self.long_desc or self.short_desc)
        )
        self.add_arguments(subparser)
        subparser.set_defaults(subcommand=self)

        subparser.add_argument(
            '-f',
            metavar="FILENAME",
            action=LoadArgsFromFile,
            help="Specify a file containing more command line arguments"
        )


    def add_arguments(self, parser: 'argparse._ActionsContainer') -> None:
        pass


    def main(self, options: 'argparse.Namespace') -> None:
        raise NotImplementedError



class ExporterSubcommand(Subcommand):
    """
    Basic PeakRDL exporter subcommand.
    Most subcommands will fall under this category as they do the following:
    - Compile one or more RDL files
    - Optionally import other non-RDL sources
    - Elaborate the register model
    - Export <something>
    """

    #: Whether this subcommand should require the user to provide an output path
    generates_output_file = True

    def __init__(self) -> None:
        self.importers = get_importer_plugins()

    def add_arguments(self, parser: 'argparse._ActionsContainer') -> None:
        compiler_arg_group = parser.add_argument_group("compilation args")
        process_input.add_rdl_compile_arguments(compiler_arg_group)
        process_input.add_elaborate_arguments(compiler_arg_group)

        process_input.add_importer_arguments(parser, self.importers)

        exporter_arg_group = parser.add_argument_group("exporter args")
        self.add_exporter_arguments(exporter_arg_group)


    def add_exporter_arguments(self, arg_group: 'argparse._ActionsContainer') -> None:
        if self.generates_output_file:
            arg_group.add_argument(
                "-o",
                dest="output",
                required=True,
                help="Output path",
            )


    def main(self, options: 'argparse.Namespace') -> None:
        rdlc = RDLCompiler()

        parameters = process_input.parse_parameters(rdlc, options.parameters)

        process_input.process_input(rdlc, self.importers, options.input_files, options)

        top = process_input.elaborate(rdlc, parameters, options)

        # Run exporter
        self.do_export(top, options)


    def do_export(self, top_node: 'AddrmapNode', options: 'argparse.Namespace') -> None:
        raise NotImplementedError



from typing import TYPE_CHECKING

from systemrdl import RDLCompiler
from systemrdl.component import Addrmap

from ..subcommand import Subcommand
from ..plugins.importer import get_importer_plugins
from .. import process_input

if TYPE_CHECKING:
    import argparse

class ListGlobals(Subcommand):
    name = "globals"
    short_desc = "list all globally accessible types that can be elaborated as top"

    def __init__(self) -> None:
        self.importers = get_importer_plugins()

    def add_arguments(self, parser: 'argparse._ActionsContainer') -> None:
        compiler_arg_group = parser.add_argument_group("compilation args")
        process_input.add_rdl_compile_arguments(compiler_arg_group)

        process_input.add_importer_arguments(parser, self.importers)

    def main(self, options: 'argparse.Namespace') -> None:
        rdlc = RDLCompiler()
        process_input.process_input(rdlc, self.importers, options.input_files, options)

        for name, comp_def in rdlc.root.comp_defs.items():
            if isinstance(comp_def, Addrmap):
                print(name)



from typing import TYPE_CHECKING, Any, List

from .import entry_points
from ..subcommand import ExporterSubcommand

if TYPE_CHECKING:
    from systemrdl.node import AddrmapNode
    import argparse


class ExporterSubcommandPluginWrapper(ExporterSubcommand):
    """
    Exporters external to this package can register a subcommand implementation
    that can be loaded into PeakRDL's subcommand list.

    The subcommand definition is provided by a class that mimics the
    members/methods of ExporterSubcommand without actually extending it

    .. code:: python

        class MyExporter:
            short_desc = "..."
            long_desc = "..."
            generates_output_file = True

            def add_exporter_arguments(self, arg_group: 'argparse.ArgumentParser') -> None:
                pass

            def do_export(self, top_node: 'AddrmapNode', options: 'argparse.Namespace') -> None:
                raise NotImplementedError
    """

    def __init__(self, name: str, exporter_plugin_cls: Any) -> None:
        super().__init__()

        self.plugin = exporter_plugin_cls()
        self.name = name
        self.short_desc = getattr(self.plugin, "short_desc")
        self.long_desc = getattr(self.plugin, "long_desc", None)
        self.generates_output_file = getattr(self.plugin, "generates_output_file", True)


    def __repr__(self) -> str:
        return "<%s '%s' from %s at 0x%x>" % (
            self.__class__.__qualname__,
            self.name,
            self.plugin.__class__,
            id(self)
        )


    def add_exporter_arguments(self, arg_group: 'argparse._ActionsContainer') -> None:
        super().add_exporter_arguments(arg_group)

        func = getattr(self.plugin, "add_exporter_arguments", None)
        if callable(func):
            func(arg_group)


    def do_export(self, top_node: 'AddrmapNode', options: 'argparse.Namespace') -> None:
        func = getattr(self.plugin, "do_export", None)
        if callable(func):
            func(top_node, options)
        else:
            raise NotImplementedError



def get_exporter_plugins() -> List[ExporterSubcommandPluginWrapper]:
    """
    Load any plugins that advertise themselves in their setup.py via the following:

    setup(
        ...
        entry_points = {
            "peakrdl.exporters": [
                'my_exporter_name = module.path.to:MyExporter'
            ]
        },
    )
    """
    exporters = []
    for ep in entry_points.get_entry_points("peakrdl.exporters"): # type: ignore
        exporter = ExporterSubcommandPluginWrapper(ep.name, ep.load())
        exporters.append(exporter)

    return exporters
